# -*- coding: utf-8 -*-

# python 2 support via python-future
from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import dict

import os

import pytest

from mariobros import mariofile

SIMPLE_MARIOFILE = """[section_one]
text one
[section_two]
text two
"""

COMPLEX_MARIOFILE = """default text

[section]    \ntext section
"""

GARBAGE_MARIOFILE = """default

[garbage_section] # garbage
"""

INVALID_SECTION_MARIOFILE = """
# spaces not allowed in section name
[section one]
"""

MORE_COMPLEX_MARIOFILE = """# default

[section_one]
text one
# comment
text two # inline comment
[section_two]
text three
[three]
[DEFAULT]
last"""


def test_parse_sections():
    simple_mariofile_sections = dict(mariofile.parse_sections(SIMPLE_MARIOFILE.splitlines(True)))
    assert len(simple_mariofile_sections) == 3
    complex_mariofile_sections = dict(mariofile.parse_sections(COMPLEX_MARIOFILE.splitlines(True)))
    assert len(complex_mariofile_sections) == 2
    assert sorted(complex_mariofile_sections.keys()) == ['DEFAULT', 'section']
    assert complex_mariofile_sections['DEFAULT'] == ['default text\n', '\n']
    with pytest.raises(mariofile.ConfigurationFileError):
        dict(mariofile.parse_sections(GARBAGE_MARIOFILE.splitlines(True)))
    with pytest.raises(mariofile.ConfigurationFileError):
        dict(mariofile.parse_sections(INVALID_SECTION_MARIOFILE.splitlines(True)))
    more_complex_mariofile_sections = dict(
            mariofile.parse_sections(MORE_COMPLEX_MARIOFILE.splitlines(True))
    )
    more_complex_mariofile_sections_keys = ['DEFAULT', 'section_one', 'section_two', 'three']
    assert sorted(more_complex_mariofile_sections.keys()) == more_complex_mariofile_sections_keys
    assert more_complex_mariofile_sections['three'] == []


CRASH_MARIOFILE_1 = '''
[a]
name
  target:
    a = 1
'''

CRASH_MARIOFILE_2 = '''
[a]
name
    variable = 1
'''


def test_statements():
    with pytest.raises(mariofile.ConfigurationFileError):
        mariofile.parse_section_body(CRASH_MARIOFILE_1.splitlines())
    with pytest.raises(mariofile.ConfigurationFileError):
        mariofile.parse_section_body(CRASH_MARIOFILE_2.splitlines())


STRING_PARSE_STATEMENTS = '''
# commento

statement

statement con commento #commento

# altro commento
'''


def test_parse_statements():
    parsed_statement = mariofile.parse_statements(STRING_PARSE_STATEMENTS.splitlines())
    assert '\n'.join(parsed_statement) == "statement\nstatement con commento"


SECTION = """
variable = 6
target: source
    task
"""

SECTION_MULTIPLE_RULE = """
target1: source1
    task1
target2: source2
    task2
"""

INVALID_CONFIG = """
not a definition
target: source
"""


def test_parse_section_body():
    output_section = {
        'action_template': '    task',
        'sources_repls': 'source',
        'variable': '6',
        'target_pattern': 'target',
    }
    assert mariofile.parse_section_body(SECTION.splitlines(True)) == output_section
    with pytest.raises(mariofile.ConfigurationFileError):
        mariofile.parse_section_body(SECTION_MULTIPLE_RULE.splitlines(True))
    with pytest.raises(mariofile.ConfigurationFileError):
        mariofile.parse_section_body(INVALID_CONFIG.splitlines(True))


INCLUDE_FILE = """
include prova.ini\t
include\taltrofile.ini

variable_definition = None
[first_section]
"""
INCLUDE_UNIQUE_FILE = "include prova.ini"


def test_parse_include():
    filepaths, current_line = mariofile.parse_include(INCLUDE_FILE.splitlines(True))
    assert filepaths == ['prova.ini', 'altrofile.ini']
    assert current_line == 4
    filepaths, current_line = mariofile.parse_include(INCLUDE_UNIQUE_FILE.splitlines(True))
    assert filepaths == ['prova.ini']
    assert current_line == 1


MARIOFILE = """[DEFAULT]
variable = 1

[section_one]
target1: source1
    task1
"""

MARIOFILE_AND_INCLUDE = """
include test_parse_config.ini

[section_include_1]
"""

MARIOFILE_INCLUDE = """
task_cmd = task_command
[section_include]
variable_include_2 = 0
target_include: source_include
\t${task_cmd}
[section_include_1]
variable_include_3 = 3
"""

TOUCH_MARIOFILE = """
DEFAULT:
    touch

[task]
target: source
    task
"""

TEST_PARSE_CONFIG = """
include test_include.ini
variable_default = 1

[section_main]
[section_include_1]
variable_include1 = 3
"""


def test_parse_config(tmpdir):
    parsed_mariofile = {
        'DEFAULT': {
            'action_template': '',
            'sources_repls': '',
            'target_pattern': '',
            'variable': '1'
        },
        'section_one': {
            'action_template': '    task1',
            'sources_repls': 'source1',
            'target_pattern': 'target1'}
    }
    mariofile.parse_config(MARIOFILE.splitlines(True)) == parsed_mariofile
    parsed_mariofile_include_test = {
        'DEFAULT': {
            'action_template': '',
            'sources_repls': '',
            'target_pattern': '',
            'task_cmd': 'task_command',
        },
        'section_include': {
            'variable_include_2': '0',
            'action_template': '\t${task_cmd}',
            'target_pattern': 'target_include',
            'sources_repls': 'source_include',
        },
        'section_include_1': {
            'action_template': '',
            'sources_repls': '',
            'target_pattern': '',
            'variable_include_3': '3',
        }
    }
    mario_folder = tmpdir.mkdir('tmpdir')
    f = mario_folder.join('test_parse_config.ini')
    f.write(MARIOFILE_INCLUDE)
    g = mario_folder.join('test_include.ini')
    g.write('')
    mario_folder.chdir()
    parsed_mariofile_include = mariofile.parse_config(
            MARIOFILE_AND_INCLUDE.splitlines(True),
            cwd=os.path.join(str(mario_folder.dirname), 'tmpdir')
    )
    for key, value in parsed_mariofile_include.items():
        assert value == parsed_mariofile_include_test[key], print(key)

    parsed_mariofile_multiple_include = {
        'DEFAULT': {
            'action_template': '',
            'sources_repls': '',
            'target_pattern': '',
            'variable_default': '1',
        },
        'section_main': {
            'action_template': u'',
            'sources_repls': u'',
            'target_pattern': u''
        },
        'section_include_1': {
            'action_template': '',
            'sources_repls': '',
            'target_pattern': '',
            'variable_include1': '3',
        }
    }
    h = mario_folder.join('test_parse_config.ini')
    h.write(TEST_PARSE_CONFIG)
    parsed_mariofile_include = mariofile.parse_config(MARIOFILE_AND_INCLUDE.splitlines(True),
                                                      cwd=os.path.join(
                                                              str(mario_folder.dirname), 'tmpdir'
                                                      ))
    assert parsed_mariofile_include == parsed_mariofile_multiple_include



# -*- coding: utf-8 -*-

# python 2 support via python-future
from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import open

import collections
import future.utils
import os
import re


class ConfigurationFileError(Exception):
    pass


def parse_statements(stream):
    """Remove blank line and trailing comments from a string.

    :param iterable stream: Mariofile stream.
    :rtype: iterable
    """
    statement = ''
    for raw_line in stream:
        # remove trailing comments, spaces and new line
        line = raw_line.partition('#')[0].rstrip()
        if re.match(r'^\s+', line):
            statement += '\n' + line
        else:
            if statement:
                yield statement
            statement = line
    if statement:
        yield statement


def parse_section_body(stream):
    """Parse the body of a section and return the section namespace.

    :param iterable stream: Mariofile stream.
    :rtype: dict
    """
    namespace = collections.OrderedDict([
        ('target_pattern', ''),
        ('sources_repls', ''),
        ('action_template', ''),
    ])
    for statement in parse_statements(stream):
        if '=' in statement.partition('\n')[0]:
            name, _, expression_template = [s.strip() for s in statement.partition('=')]
            namespace[name] = expression_template
        elif ':' in statement.partition('\n')[0]:
            if namespace['target_pattern']:
                raise ConfigurationFileError("Section must have only one rule")
            target_sources, _, action_template = statement.strip().partition('\n')
            namespace['action_template'] = action_template
            # FIXME: you cannot have SOURCES path with ":" inside, like "s3://".
            target_pattern, _, sources = target_sources.rpartition(':')
            namespace['sources_repls'] = sources.strip()
            namespace['target_pattern'] = target_pattern.strip()
        else:
            raise ConfigurationFileError("Statement not an assignment nor a rule: %r" % statement)
    return namespace


def parse_sections(stream):
    """Parse a Mariofile and return a dictionary with the section name as the key and the section
    content as the value.

    :param iterable stream: Mariofile stream.
    :rtype: tuple
    """
    section, body_stream = 'DEFAULT', []
    for line_number, line in enumerate(stream, 1):
        match = re.match(r'^\[(.*)\](.*)$', line)
        if match:
            section_name = match.group(1).strip()
            garbage = match.group(2).strip()
            if not future.utils.isidentifier(section_name):
                raise ConfigurationFileError(
                    "Invalid section name at line %r: %r" % (line_number, match.group(0))
                )
            if garbage:
                raise ConfigurationFileError(
                    "Garbage after section name at line %r: %r" % (line_number, match.group(0))
                )
            else:
                yield section, body_stream
                section = match.group(1)
                body_stream = []
        else:
            body_stream += [line]
    yield section, body_stream


def parse_include(raw_lines):
    """Parse the heading of the raw_lines and return a list of filepaths to be included.

    :param iterable raw_lines: Mariofile raw_lines.
    :rtype: (list, iterable)
    """
    filepaths = []
    current_line = 0
    # workaround for files with only includes and without newline at the end of the file.
    raw_lines += '\n'
    for current_line, raw_line in enumerate(raw_lines):
        line = raw_line.partition('#')[0].rstrip()
        include_match = re.match(r'^include\s+(.*)\s*$', line)
        if not line.strip():
            continue
        elif include_match:
            filepaths.append(include_match.group(1))
        else:
            break
    return filepaths, current_line


def parse_config(raw_lines, cwd='.'):
    """Parse the input Mariofile and the included Mariofiles recursively and return a dictionary
    with the section name as the key and the section namespace as the value.

    :param iterable raw_lines: Mariofile per lines.
    :param str cwd: Current working directory.
    :rtype: dict
    """
    include_paths, end_include_line = parse_include(raw_lines)
    section_namespaces = collections.OrderedDict()
    for include_path in include_paths:
        for section, section_namespace in parse_mariofile(include_path, cwd=cwd).items():
            namespace = section_namespaces.setdefault(section, collections.OrderedDict())
            namespace.update(section_namespace)
    for section, body_stream in parse_sections(raw_lines[end_include_line:]):
        namespace = section_namespaces.setdefault(section, collections.OrderedDict())
        namespace.update(parse_section_body(body_stream))
    return section_namespaces


def parse_mariofile(file_path, cwd='.'):
    """Parse input mariofile.

    :param unicode file_path: File path.
    :param str cwd: Current working directory.
    :return: dict
    """
    file_path = os.path.join(cwd, file_path)
    file_cwd = os.path.dirname(file_path)
    with open(file_path) as stream:
        return parse_config(stream.read().splitlines(True), cwd=file_cwd)



# -*- coding: utf-8 -*-

# python 2 support via python-future
from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import bytes, dict, int, str

import atexit
import collections
import distutils.spawn
import importlib
import logging
import re
import shlex
import subprocess
import sys
import uuid

import future.utils
import luigi
from luigi.contrib.s3 import S3Target
import mako.template


LOGGER = logging.getLogger('luigi-interface')

TEMPLATE = """% for var_def, val_def in default_namespace.items():
    % if var_def not in ['action_template', 'sources_repls', 'target_pattern']:
${var_def} = ${val_def}
    %endif
% endfor

% if default_namespace['target_pattern']:
${default_namespace['target_pattern']}: ${default_namespace['sources_repls']}
    ${default_namespace['action_template']}
% endif
% for task_name, section_namespace in section_namespaces.items():
    % if task_name != 'DEFAULT':
[${task_name}]
        %for var, val in section_namespaces[task_name].items():
            % if var not in ['action_template', 'sources_repls', 'target_pattern']:
${var} = ${val}
            % endif
        % endfor
${section_namespace['target_pattern']}: ${section_namespace['sources_repls']}
    ${section_namespace['action_template']}
    % endif

% endfor
"""


def pretty_unicode(obj):
    """Filter to pretty print iterables."""
    if not isinstance(obj, (str, bytes)):
        try:
            return ' '.join(str(item) for item in obj)
        except TypeError:
            pass
    return str(obj)


class ExistingFile(luigi.ExternalTask):
    """Define Luigi External Task class for existing files requires."""

    target = luigi.Parameter()

    def output(self):
        return luigi.LocalTarget(self.target)


class ReRuleTask(luigi.Task):
    """Define Luigi task class through regular expression.
    """
    # target_pattern = ''
    # sources_repls = []

    # action_namespace = {}
    # action_template = ''
    # SHALL = '/bin/bash'

    @staticmethod
    def factory(
            name, target_pattern, sources_repls=(), action_template='', action_namespace={},
            priority=0, worker_timeout=None, resources={}, disabled=False, dry_run_suffix='',
            SHELL='/bin/bash'):
        """Create Luigi task class.

        :param str name: Task name.
        :param str target_pattern: Target pattern.
        :param list sources_repls: List of source replacements.
        :param str action_template: Action template.
        :param dict action_namespace: Action namespace.
        :param int priority: Priority Luigi task metadata.
        :param int worker_timeout: Worker timeout Luigi task metadata.
        :param dict resources: Resources Luigi task metadata.
        :param bool disabled: Disabled Luigi task metadata.
        :param unicode dry_run_suffix: Suffix to be added to file created during dry run.
        :rtype: subclass_of_ReRuleTask
        """
        # FIXME: move class init code to init method?
        if not target_pattern.startswith('(?<![^/])'):
            target_pattern = '(?<![^/])' + target_pattern
        if not target_pattern.endswith('$'):
            target_pattern += '$'
        if action_template.strip() == '':
            action_template = 'echo "${SOURCES} -> ${TARGET}"'
        _target_pattern = re.compile(target_pattern)
        return type(future.utils.native_str(name), (ReRuleTask,), locals())

    @classmethod
    def match(cls, target):
        """Perform target matching.

        :rtype: bool
        """
        return bool(cls._target_pattern.search(target))

    target = luigi.Parameter()

    def render_sources(self):
        """Perform rendering of the sources.

        :rtype: str
        """
        return tuple(self._target_pattern.sub(repl, self.target) for repl in self.sources_repls)

    def render_action(self):
        """Perform rendering of the action.

        :rtype: str
        """
        sources = self.render_sources()
        match = self._target_pattern.match(self.target)
        target_namespace = dict(TARGET=self.target, SOURCES=sources, MATCH=match)
        return render_template(
                self.action_template, target_namespace, default_namespace=self.action_namespace
        )

    def render_shell(self):
        sources = self.render_sources()
        match = self._target_pattern.match(self.target)
        target_namespace = dict(TARGET=self.target, SOURCES=sources, MATCH=match)
        return render_template(
                self.SHELL, target_namespace, default_namespace=self.action_namespace
        )

    def output(self):
        """
        The output that this Task produces.

        See :ref:`Task.output`
        :rtype: luigi.LocalTarget
        """
        if self.target.startswith('s3://'):
            return S3Target(self.target)
        else:
            return luigi.LocalTarget(self.target + self.dry_run_suffix)

    def requires(self):
        """
        The Tasks that this Task depends on.

        See :ref:`Task.requires`
        :rtype: list
        """
        required = []
        for source in self.render_sources():
            for task_rule in ReRuleTask.__subclasses__():
                if task_rule.match(source):
                    required.append(task_rule(target=source))
                    break
            else:
                required.append(ExistingFile(source))
        return required

    def run(self):
        """
        The task run method, to be overridden in a subclass.

        See :ref:`Task.run`
        """
        action = self.render_action()
        if self.dry_run_suffix:
            # log intended command line but touch the dry_run target instead
            LOGGER.info(action)
            action = 'touch ' + self.target + self.dry_run_suffix
            # register the dry_run target removal at program exit
            atexit.register(self.remove_dry_run_file)
            args = ['/bin/bash', '-c', action]
        else:
            shell = self.render_shell()
            args = shlex.split(shell) + ['-c', action]
            # be sure to use the abspath of the executable based on the PATH environment variable
            args[0] = distutils.spawn.find_executable(args[0])
        LOGGER.info('COMMAND: {}'.format(args))
        subprocess.check_call(args)

    def remove_dry_run_file(self):
        """Remove files generated by dry run process."""
        subprocess.call('rm -f ' + self.target + self.dry_run_suffix, shell=True)


def render_template(template, local_namespace, default_namespace={}):
    """Return the rendered template merging local and default namespaces.

    :param unicode template: Template.
    :param dict local_namespace: Local namespace.
    :param dict default_namespace: Default namespace.
    :rtype: str
    """
    namespace = default_namespace.copy()
    namespace.update(local_namespace)
    if 'IMPORT_MODULES' in namespace:
        import_modules = namespace['IMPORT_MODULES'].split()
        namespace.update({name: importlib.import_module(name) for name in import_modules})

    template_object = mako.template.Template(
        template,
        strict_undefined=True,
        imports=['from mariobros.mario import pretty_unicode'],  # enable the filter
        default_filters=['pretty_unicode'],
    )
    return template_object.render(**namespace)


def render_namespace(namespace, default_namespace={}, skip_names=('action_template', 'SHELL')):
    """Return Render section namespaces with default section namespaces also.

    :param dict namespace: Section namespace.
    :param dict default_namespace: default section namespace.
    :param list skip_names: Namespace names to skip in the render process.
    :rtype: dict
    """
    torender_namespace = {k: v for k, v in namespace.items() if k not in skip_names}
    rendered_namespace = {k: v for k, v in namespace.items() if k in skip_names}
    while len(torender_namespace):
        loop = True
        for key, value_template in list(torender_namespace.items()):
            try:
                value = render_template(value_template, rendered_namespace, default_namespace)
                torender_namespace.pop(key)
                rendered_namespace[key] = value
                loop = False
            except NameError:
                pass
        if loop:
            raise NameError("Can't render: {!r}".format(torender_namespace))
    return collections.OrderedDict((k, rendered_namespace[k]) for k in namespace)


def register_tasks(namespaces, default_namespace={}, dry_run_suffix=''):
    """Return a Luigi task class after parsed Luigi task metadata.

    :param dict namespaces: Task namespaces.
    :param dict default_namespace: Default namespaces.
    :param unicode dry_run_suffix: Suffix to be added to file created during dry run.
    :rtype: iterable
    """

    for task_name, namespace in namespaces.items():
        action_namespace = default_namespace.copy()
        action_namespace.update(namespace)
        task_keys = ['target_pattern', 'sources_repls', 'action_template', 'SHELL']
        task_namespace = {k: action_namespace[k] for k in task_keys if k in action_namespace}
        task_namespace['sources_repls'] = task_namespace['sources_repls'].split()
        # luigi attributes
        task_namespace['resources'] = {k.partition('_')[2]: int(v) for k, v in namespace.items()
                                       if k.startswith('RESOURCES_')}
        task_namespace.update(
            {k: int(namespace[k]) for k in ['priority', 'disabled', 'worker_timeout']
             if k in namespace})
        yield ReRuleTask.factory(
                task_name, dry_run_suffix=dry_run_suffix, action_namespace=action_namespace,
                **task_namespace
        )


def print_namespaces(default_namespace, section_namespaces):
    """Print namespaces with the MarioFile format.

    :param dict default_namespace: Default namespace dictionary.
    :param dict section_namespaces: Section namespaces dictionary.
    :return: str
    """
    template = mako.template.Template(TEMPLATE)
    namespaces = template.render(
            default_namespace=default_namespace, section_namespaces=section_namespaces
    )
    return namespaces


def render_config(section_namespaces):
    """Parse and render a MarioFile.

    :param dict section_namespaces: Section namespaces dictionary.
    :return: (dict, dict, dict)
    """
    default_namespace = render_namespace(section_namespaces['DEFAULT'])
    rendered_namespaces = collections.OrderedDict(
            (k, render_namespace(v, default_namespace)) for k, v in section_namespaces.items()
    )
    return default_namespace, rendered_namespaces


def mario(rendered_namespaces, default_namespace, targets=('DEFAULT',), dry_run=False):
    """Generate Luigi tasks' file from MarioFile and Luigi template file

    :param dict rendered_namespaces: Rendered namespaces dictionary.
    :param dict default_namespace: Default namespace dictionary.
    :param iterable targets: List of targets.
    :param bool dry_run: Dry run flag.
    :rtype : iterable
    """
    # ensure '.' is present in sys.path so 'IMPORT_MODULES = local_module' works
    if '.' not in sys.path:
        sys.path.append('.')

    dry_run_suffix = '-dry_run-' + str(uuid.uuid4()) if dry_run else ''
    rendered_namespaces = collections.OrderedDict(reversed(list(rendered_namespaces.items())))
    tasks = list(register_tasks(
            rendered_namespaces, default_namespace=default_namespace, dry_run_suffix=dry_run_suffix
    ))
    target_tasks = []
    for target in targets:
        for task_rule in tasks:
            if task_rule.match(target):
                target_tasks.append(task_rule(target=target))
                break
    return target_tasks




# python 2 support via python-future
from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import str

import glob
import os
import shlex

import click.testing
import pytest

from mariobros import cli

TOUCH_FILE = r"""touched-11-22.file: touched-11.file touched-22.file
    touch ${TARGET}

[touch_single]
touched-([^-.]+).file:
    touch ${TARGET}"""


def test_mariobros(tmpdir):
    mario_folder = tmpdir.mkdir('tmpdir')
    mario_folder.chdir()
    f = mario_folder.join('touch_file.ini')
    f.write(TOUCH_FILE)
    cli.mariobros(
            targets=['touched-11-22.file'], mariofile='touch_file.ini', local_scheduler=True
    )
    created_files = sorted(glob.glob('touched*'))
    assert created_files == ['touched-11-22.file', 'touched-11.file', 'touched-22.file']
    cli.mariobros(
            targets=['touched-11-22.file', ], mariofile='touch_file.ini', local_scheduler=True,
            dry_run=True, workers=2
    )
    cli.mariobros(
            targets=['touched-11-22.file', ], mariofile='touch_file.ini', local_scheduler=True,
            print_ns=True
    )

    with pytest.raises(AssertionError):
        cli.mariobros(targets=['touched-11-22.file'.encode('utf-8')])


def test_cli(tmpdir):
    mario_folder = tmpdir.mkdir('tmpdir')
    mario_folder.chdir()
    f = mario_folder.join('touch_file.ini')
    f.write(TOUCH_FILE)

    runner = click.testing.CliRunner()
    mariofile = os.path.join(mario_folder.dirname, 'tmpdir/touch_file.ini')
    args = shlex.split('-f {} --local-scheduler touched-11-22.file'.format(mariofile))
    # split casts the args to str in python2
    args = [str(a) for a in args]
    result = runner.invoke(cli.main, args=args)
    assert not result.exception
    args = shlex.split('-f {} --local-scheduler'.format(mariofile))
    result = runner.invoke(cli.main, args=args)
    assert not result.exception



# -*- coding: utf-8 -*-

# python 2 support via python-future
from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import next

import collections
import os

import luigi
import pytest

from mariobros import mario
from mariobros import mariofile


cw_dir = os.path.dirname(__file__)


def test_pretty_unicode():
    assert mario.pretty_unicode(range(3)) == '0 1 2'
    assert mario.pretty_unicode('test') == 'test'


class test_ExistingFile():
    existing_file = mario.ExistingFile(__file__)
    assert existing_file.output().exists()


def test_render_template():
    template = '${global_name}-${local_name}'
    global_ns = {'global_name': 'GLOBAL'}
    local_ns = {'local_name': 'LOCAL'}
    rendered = mario.render_template(template, global_ns, local_ns)
    assert rendered == 'GLOBAL-LOCAL'


def test_render_namespace():
    assert mario.render_namespace({'key': 'value'}) == {'key': 'value'}
    assert mario.render_namespace({'key': '${value}'}, {'value': '3'}) == {'key': '3'}
    small_namespace = collections.OrderedDict({'key': '${value}', 'value': '3'})
    assert mario.render_namespace(small_namespace) == {'key': '3', 'value': '3'}
    big_namespace = collections.OrderedDict({'key': '${value}', 'value': '${meta}', 'meta': '3'})
    assert mario.render_namespace(big_namespace) == {'key': '3', 'value': '3', 'meta': '3'}
    with pytest.raises(NameError):
        mario.render_namespace({'a': '${b}'})
    with pytest.raises(NameError):
        mario.render_namespace({'a': '${b}', 'b': '${a}'})
    assert mario.render_namespace({'a': '${b}'}, skip_names=['a']) == {'a': '${b}'}


NAMESPACES = collections.OrderedDict([
    ('rule1', {
        'target_pattern': '(.*).target',
        'sources_repls': '\\1.source',
        'action_template': 'action -o ${TARGET} ${SOURCES}',
    }),
    ('rule2', {
        'RESOURCES_cpu': 2,
        'RESOURCES_mem': 4,
        'priority': 20,
        'target_pattern': '(.*).source',
        'sources_repls': '\\1.orig',
        'action_template': '',
    }),
])


def test_ReRuleMixin():
    dry_run_suffix = '-test'
    task_rules = list(mario.register_tasks(NAMESPACES, dry_run_suffix=dry_run_suffix))

    task_rule1 = task_rules[0]
    assert task_rule1.match('file.target')
    assert not task_rule1.match('wrong')
    task1 = task_rule1(target='file.target')
    assert task1.render_sources() == ('file.source',)
    assert task1.render_action() == 'action -o file.target file.source'

    output = task1.output()
    assert isinstance(output, luigi.Target)
    assert output.path == 'file.target' + dry_run_suffix

    requires = task1.requires()
    assert len(requires) == 1
    assert requires[0].target == 'file.source'

    task1.run()

    task_rule2 = task_rules[1]
    task2 = task_rule2(target=requires[0].target)
    requires2 = task2.requires()
    assert len(requires2) == 1
    assert requires2[0].target == 'file.orig'

    task2.remove_dry_run_file()


def test_register_tasks():
    task_rules = mario.register_tasks(NAMESPACES)

    task_rule1 = next(task_rules)
    assert task_rule1.target_pattern == '(?<![^/])(.*).target$'

    task_rule2 = next(task_rules)
    assert task_rule2.resources == {'cpu': 2, 'mem': 4}
    assert task_rule2.priority == 20
    assert task_rule2.match('file.source')


TOUCH_MARIOFILE = """
DEFAULT:
    touch

[task]
target: source
    task
"""


def test_print_namespaces():
    section_namespaces = mariofile.parse_config(TOUCH_MARIOFILE.splitlines(True))
    default_namespace = mario.render_namespace(section_namespaces['DEFAULT'])
    output = """
DEFAULT:\x20
        touch
[task]
target: source
        task
"""
    # workaround for different behaviour of mako in cpython2, cpython3, pypy2 and pypy3
    assert mario.print_namespaces(
            default_namespace, section_namespaces
    ).replace('\n\n', '\n').replace('\n\n', '\n') == output


def test_render_config(tmpdir):
    mario_folder = tmpdir.mkdir('tmpdir')
    f = mario_folder.join('render_config.ini')
    f.write(TOUCH_MARIOFILE)
    mario_folder.chdir()
    section_namespaces = mariofile.parse_mariofile('render_config.ini')
    default_namespace, rendered_namespaces = mario.render_config(section_namespaces)
    assert section_namespaces['DEFAULT']['action_template'] == '    touch'
    assert default_namespace['action_template'] == '    touch'
    assert rendered_namespaces['DEFAULT']['action_template'] == '    touch'


def test_mario(tmpdir):
    mario_folder = tmpdir.mkdir('tmpdir')
    f = mario_folder.join('test_mario.ini')
    f.write(TOUCH_MARIOFILE)
    mario_folder.chdir()
    section_namespaces = mariofile.parse_mariofile('test_mario.ini')
    default_namespace, rendered_namespaces = mario.render_config(section_namespaces)
    touch_tasks = mario.mario(rendered_namespaces, default_namespace)
    assert touch_tasks[0].target == 'DEFAULT'
    touch_tasks = mario.mario(rendered_namespaces, default_namespace, targets=['target'])
    assert touch_tasks[0].target == 'target'
    assert touch_tasks[0].name == 'task'



# -*- coding: utf-8 -*-

# python 2 support via python-future
from __future__ import absolute_import, division, print_function, unicode_literals
from builtins import int, str

import click
import luigi

from mariobros import mario
from mariobros import mariofile as mariofile_


def mariobros(
        targets=('DEFAULT',), mariofile='mario.ini', print_ns=False, dry_run=False, workers=1,
        port=8082,
        **kwargs
):
    """Main mariobros entry point. Parse the configuration file and launch the build of targets.

    :param sequence targets: List of targets.
    :param unicode mariofile: MarioFile name.
    :param bool print_ns: Flag to print namespace.
    :param bool dry_run: Dry run flag.
    :param int workers: Number of workers.
    :param dict kwargs: Passed to the luigi.build function.
    """
    assert all(isinstance(target, str) for target in targets)

    if dry_run and workers > 1:
        workers = 1
        mario.LOGGER.warning('Dry run is incompatible with multiprocessing. Setting --workers=1')
    section_namespaces = mariofile_.parse_mariofile(mariofile)
    default_namespace, rendered_namespaces = mario.render_config(section_namespaces)
    if print_ns:
        namespaces = mario.print_namespaces(default_namespace, section_namespaces)
        print(namespaces)
    else:
        target_tasks = mario.mario(
            rendered_namespaces, default_namespace, targets=targets, dry_run=dry_run
        )
        luigi.build(target_tasks, workers=workers, scheduler_port=port, **kwargs)


@click.command()
@click.argument('targets', nargs=-1, type=click.Path())
@click.option(
    '--mariofile', '--file', '-f', default='mario.ini',
    help="Main configuration file", type=click.Path(exists=True, dir_okay=False))
@click.option('--port', '-p', default=8082, help="Set `luigi.build` scheduler_port parameter.")
@click.option('--workers', default=1, help="Set the number of workers", type=int)
@click.option('--local-scheduler', is_flag=True, help="Run local scheduler.")
@click.option('--print-ns', is_flag=True,
              help="Print namespaces: Print the MarioFile with the included tasks and variables")
@click.option(
    '--dry-run', '-n', is_flag=True,
    help="Don't actually run any commands; just print them.")
def main(targets, **kwargs):
    if not targets:
        targets = ('DEFAULT',)
    else:
        targets = [str(token) for token in targets]
    mariobros(targets, **kwargs)




from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

version = '0.4.3'

setup(
    name='mariobros',
    version=version,
    description='Simple configuration for Spotify Luigi.',
    long_description=long_description,
    url='https://github.com/bopen/mariobros',
    download_url='https://github.com/bopen/mariobros/archive/%s.tar.gz' % version,
    author='B-Open Solutions srl',
    author_email='oss@bopen.eu',
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: System :: Monitoring',
    ],
    packages=find_packages(exclude=['contrib', 'docs']),
    zip_safe=False,
    install_requires=[
        'click',
        'mako',
        'future',
        'luigi',
        'sqlalchemy',
    ],
    extras_require={
        'dev': [
            'check-manifest',
            'pytest',
            'pytest-cov',
            'pytest-flakes',
            'pytest-pep8',
            'pytest-pylint',
            'tox',
            'sphinx',
            'sphinx_rtd_theme',
            'ipython',
            'ipdb',
        ],
    },
    entry_points={
        'console_scripts': [
            'mario=mariobros.cli:main',
        ],
    },
)
